/**
 * Created by TFG on 07/15/2020.
 * Management for the Master Item loading data
 *
 * CAUTION ! August 2020 note !
 * Since Clinic Profile process is not complete, in order to have an intermediate release CP will ignore any record
 * coming in the queue !
 * See function processMetadata - set abort flag to true !
 * That flag should be set to false for normal working !
 */
var app = require("../../child");

var LoaderHelper = require("../helpers/LoaderHelper");
var awcblcommon = require('../config/awcblcommon');

var currentTransaction;
var clinicCatalogs = [];    // List of clinic catalog records associated to the input record
// Doing that because unable to assign correctly the models variable where needed
var clinicCatalog = {
    clinicId: "",
    catalogId: ""
};

// Object to deal with metadata
var CPmetadata = {
    transactionId: "N/A",     //    - transaction ID
    fileName: "--",           //    - name of the originating CSV file
    expectedNumRec: 0,        //    - number to record to be received (remove 1 when counting)
    rcvCounter: 0,            //    - counter of record received  (remember to add 1 when showing the value - counter from 0 not 1)
    dbCounter: 0,             //    - counter of record saved on DB ((remember to add 1 when showing the value - counter from 0 not 1)
    endLoad: false,           //    - flag to indicate the current load as ended
    endProc: false,           //    - flag to indicate the current processing as ended
    abort: false,             //    - flag to abort the transaction ! If true all the incoming records will ack but not processed
    sendEndLoad: true,        //    - flag to be used to send out only ONE log notification about end loading
    sendEndProc: true,        //    - flag to be used to send out only ONE log notification about end processing
    sendAbort: true,          //    - flag to be used to send out only ONE log notification about abort
    sendEndAssociate: true,   //    - flag to be used to send out only ONE log notification about end associate saving
    incrementRcvCounter : function() {
        this.rcvCounter++;
        if (this.rcvCounter >= (this.expectedNumRec -1)) {
            this.endLoad = true;
        }
    },
    incrementDbCounter : function() {
        this.dbCounter++;
        // Here the test is little bit different because the position of the incrementDbCounter
        // Starting from zero, when process the number 9 (the latest) it will be incremented by one and
        // then the test
        if (this.dbCounter >= this.expectedNumRec) {
            this.endProc = true;
        }
    }
};

// Create array of metadata objects
var metadata = [ CPmetadata ];

module.exports = function(winston, rabbit, models, config, cb) {
    // variable to hold starting time
    var started = Date.now();

    // expected message count
    var logger = winston.loggers.get("main");

    // always setup your message handlers first
    rabbit.nackOnError();

    // this handler will handle messages sent from the publisher
    rabbit.handle({
        queue: 'topic-fresenius-loader-cpq',
        type: 'loader.cp'
    }, function (msg) {
            logger.debug(process.pid + ":LOADER CP POST Received:",
            JSON.stringify(msg.body,null, 4),
            "routingKey:", msg.fields.routingKey);
        var helper = new LoaderHelper(models, winston);

        // Metadata management - use the currentTransaction to monitor the load
        currentTransaction = processMetadata(winston, msg.body);

        // Intercept wrong incoming records !
        if (currentTransaction === null ||
            currentTransaction === undefined) {
            // Something wrong handling metadata !!!
            // Ignore incoming record !
            msg.ack();      // Acknowledge the message in the queue to clean it up
        } else {

            if (currentTransaction.abort === true) {
                // Something wrong handling metadata !!!
                // Abort upload !!!

                if (currentTransaction.sendAbort === true) {
                    // Write log to notify successfull end of operation
                    awcblcommon.writeAwcblEntryLog(app,
                        currentTransaction.fileName,
                        currentTransaction.transactionId,
                        "BLOCKED",
                        "Transaction ignored - metadata error").then(response => {
                        logger.debug(response);
                    }, reject => {
                        logger.error(reject);
                    });
                    currentTransaction.sendAbort = false;     // Disable sending other LOG messages about abort
                }
                msg.ack();
            } else {
                // NOT ABORT !! Process the incoming record !

                // Extract and prepare the data to be saved at the end of loading
                processAssociatedData(winston, msg.body);

                // Store the record on the DB via helper
                helper.saveLoaderCP(msg.body).then(response => {
                    // Response - last record saved
                    logger.debug("Record " + currentTransaction.dbCounter + " saved : " + response);
                    currentTransaction.incrementDbCounter();        // Increment saved record counter

                    if (currentTransaction.endProc === true && currentTransaction.sendEndProc === true) {
                        // End of process saving main table !!
                        // This is not YET the end of processing since the tbl_clinic_catalog needs now to be updated !

                        // Write log to notify successfull end of main processing
                        awcblcommon.writeAwcblEntryLog(app,
                            currentTransaction.fileName,
                            currentTransaction.transactionId,
                            "PROCESSING",
                            "Processed " + currentTransaction.dbCounter + "  records").then(response => {
                            logger.debug(response);
                        }, reject => {
                            logger.error(reject);
                        });

                        // Call here the function to save the data in the such table !
                        helper.saveLoaderCPassociated(clinicCatalogs).then(response => {
                            // Response - updated tbl_clinic_catalog table

                            if (currentTransaction.sendEndAssociate === true) {
                                currentTransaction.sendEndAssociate = false;

                                logger.debug("Associated Clinic Catalog saving SUCCESS : " + response);

                                // Write log to notify successfull end of operation
                                awcblcommon.writeAwcblEntryLog(app,
                                    currentTransaction.fileName,
                                    currentTransaction.transactionId,
                                    "PROCESSED",
                                    "Updated Clinic Catalog table").then(response => {
                                    logger.debug(response);
                                }, reject => {
                                    logger.error(reject);
                                });
                            }
                        }, reject => {
                            logger.error("Error from helper for CP associated : " + reject);

                            // Write log to notify error happened
                            awcblcommon.writeAwcblEntryLog(app,
                                currentTransaction.fileName,
                                currentTransaction.transactionId,
                                "ERROR",
                                "Error updating associated : " + reject + " - Contact assistance").then(response => {
                                logger.debug(response);
                            }, reject => {
                                logger.error(reject);
                            });

                        }, reject => {
                            logger.error(reject);
                        });
                        currentTransaction.sendEndProc = false;     // Disable sending other LOG messages about ending process
                    }
                    msg.ack();
                }, reject => {
                    // Record not saved ! Ack message anyway to remove from the queue
                    msg.ack();
                    logger.error("Error from helper for CP main : " + reject);

                    // Write log to notify error happened
                    awcblcommon.writeAwcblEntryLog(app,
                        currentTransaction.fileName,
                        currentTransaction.transactionId,
                        "ERROR",
                        "Error loading : " + reject + " - Contact assistance").then(response => {
                        logger.debug(response);
                    }, reject => {
                        logger.error(reject);
                    });
                });

                // Shows in the log statistics about processing
                logger.info("Transaction " + currentTransaction.transactionId +
                    " - expected " + currentTransaction.expectedNumRec +
                    " - received " + currentTransaction.rcvCounter +
                    " - stored " + currentTransaction.dbCounter +
                    " - end load : " + currentTransaction.endLoad.toString() +
                    " - end proc : " + currentTransaction.endProc.toString());

                if (currentTransaction.endLoad === true && currentTransaction.sendEndLoad === true) {
                    logger.info("Transaction " + currentTransaction.transactionId + " ENDED !");

                    // Write log to notify successfull end of operation
                    awcblcommon.writeAwcblEntryLog(app,
                        currentTransaction.fileName,
                        currentTransaction.transactionId,
                        "LOADED",
                        "Loaded " + (currentTransaction.rcvCounter + 1) + " records").then(response => {
                        logger.debug(response);
                    }, reject => {
                        logger.error(reject);
                    });
                    currentTransaction.sendEndLoad = false;     // Disable sending other LOG messages about ending loading
                }
            }
        }
    });

    logger.info(process.pid + ':Set up loader cpq OK');
    return require("../config/RabbitTopology.js")(rabbit, "loader.cp", config);
};

// Function to handle the metadata related to a load.
// In case of errors, the function will return an object with the abort flag set to true
function processMetadata( winston, msg) {
    var path = require('path');
    var logger = winston.loggers.get("main");

    var startingCSVindex = 3;
    var inTransactionId;
    var inFileName;
    var inExpectedNumRec;

    // Extract metadata from input message - it is expected :
    //   csv_line[0] -> transactionId
    //   csv_line[1] -> file name
    //   csv_line[2] -> expected number of messages
    var csv_line = msg.split(/,(?=(?:(?:[^"]*"){2})*[^"]*$)/);
    csv_line = csv_line || [];

    // Retrieve metadata
    inTransactionId = csv_line[0].split(',').join('');
    inFileName = path.basename(csv_line[1].split(',').join(''), ".csv") + ".csv";
    inExpectedNumRec = parseInt(csv_line[2].split(',').join(''), 10);

    // Check if the transaction ID exists on the array
    let currentTransaction = metadata.find(CPmetadata => CPmetadata.transactionId === inTransactionId);

    if (currentTransaction !== undefined) {
        currentTransaction.incrementRcvCounter();
    } else {
        logger.debug("New transaction identified : " + inTransactionId);

        // Not found object - add one
        CPmetadata.transactionId    = inTransactionId;
        CPmetadata.fileName         = inFileName;
        CPmetadata.expectedNumRec   = inExpectedNumRec;
        CPmetadata.rcvCounter       = 0;
        CPmetadata.dbCounter        = 0;
        CPmetadata.endLoad          = false;
        CPmetadata.endProc          = false;
        CPmetadata.sendEndLoad      = true;
        CPmetadata.sendEndProc      = true;
        CPmetadata.sendAbort        = true;
        CPmetadata.abort            = true; // TEMP !!! Force true to ignore incoming records for CP !!!
        CPmetadata.sendendAssociate = true;

        metadata.push(CPmetadata);
        currentTransaction = metadata.find(CPmetadata => CPmetadata.transactionId === inTransactionId);
    }

    return currentTransaction;
}

// Function to handle the CP associated data to store after saving all the data
function processAssociatedData( winston, msg) {
    var path = require('path');
    var logger = winston.loggers.get("main");

    var startingCSVindex = 3;
    var inClinicId;
    var inTestCatalog;

    // Extract metadata from input message - it is expected :
    //   csv_line[0] -> transactionId
    //   csv_line[1] -> file name
    //   csv_line[2] -> expected number of messages
    var csv_line = msg.split(/,(?=(?:(?:[^"]*"){2})*[^"]*$)/);
    csv_line = csv_line || [];

    // Retrieve catalog related data !!
    inClinicId = csv_line[startingCSVindex].split(',').join('');
    inTestCatalog = csv_line[startingCSVindex+1].split('"').join('');

    logger.debug("processAssociatedData - csv_line[3] = " + inClinicId);
    logger.debug("                      - csv_line[4] = " + inTestCatalog);

    // Prepare the list of the record for tbl_clinic_catalog to add
    var catIds = inTestCatalog.split("#").join('');
    logger.debug("Found " + catIds.length + " catalogs ID");

    for (var fn = 0; fn < catIds.length; fn++) {
        clinicCatalog.clinicId = inClinicId;
        clinicCatalog.catalogId = catIds[fn];
        clinicCatalogs.push(clinicCatalog);

        logger.debug(" Pushed : " + JSON.stringify(clinicCatalog,null, 4));
    }
}
